{-|
Module      : Data.Graph.Clustering.FLouvain
Description : Purely functional (Inductive) Louvain clustering
Copyright   : (c) Alexandre Delanoë, CNRS, 2020-Present
License     : AGPL + CECILL v3
Maintainer  : alexandre.delanoe+louvain@iscpif.fr
Stability   : experimental
Portability : POSIX

# Louvain Algorithm

Our algorithm is divided into two phases that are repeated iteratively.
Assume that we start with a weighted network of N nodes. First, we
assign a different community to each node of the network. So, in this
initial partition there are as many communities as there are nodes.
Then, for each node i we consider the neighbours j of i and we evaluate
the gain of modularity that would take place by removing i from its
community and by placing it in the community of j. The node i is then
placed in the community for which this gain is maximum (in the case of
a tie we use a breaking rule), but only if this gain is positive. If
no positive gain is possible, i stays in its original community. This
process is applied repeatedly and sequentially for all nodes until
no further improvement can be achieved and the first phase is then
complete.

Reference:
Blondel, Vincent D; Guillaume, Jean-Loup; Lambiotte, Renaud;
Lefebvre, Etienne (9 October 2008). "Fast unfolding of communities
in large networks". Journal of Statistical Mechanics: Theory and
Experiment. 2008 (10): P10008. arXiv:0803.0476 Freely accessible.
doi:10.1088/1742-5468/2008/10/P10008.

-}

{-# LANGUAGE ConstrainedClassMethods #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE InstanceSigs      #-}
{-# LANGUAGE NoImplicitPrelude #-}

module Data.Graph.Clustering.FLouvain
  where

import Protolude
import Data.Graph.Inductive
import qualified Data.List as DL
import Data.Tuple.Extra (fst3)

--  "glue" : function to gather/merge communities
--  "klue" : function to split communities
data ClusteringMethod = Glue | Klue
  deriving (Eq)


------------------------------------------------------------------------
-- | Specific FGL needed functions

-- | Find LNode of a node (i.e. a node with label)
lnode :: (Graph gr) => gr a b -> Node -> Maybe (LNode a)
lnode cgr n = case lab cgr n of
  Nothing -> Nothing
  Just l  -> Just (n, l)

-- | Fold over graph definitions: type and function

-- | Fold over graph type
type CFunFold a b c = Context a b -> c -> c

-- | Fold over graph function
xdfsFoldWith :: (Graph gr)
    => CFun     a b [Node]
    -> CFunFold a b c
    -> c
    -> [Node]
    -> gr a b
    -> c
xdfsFoldWith _ _ acc [] _            = acc
xdfsFoldWith _ _ acc _ g | isEmpty g = acc
xdfsFoldWith d f acc (v:vs) g =
  case match v g of
    (Just c, g')  -> xdfsFoldWith d f (f c acc) (d c++vs) g'
    (Nothing, g') -> xdfsFoldWith d f acc vs g'

------------------------------------------------------------------------

-- Our basic graph. Nodes have custom labels. Edges have weight assigned to them.
type FEdge = Double
type FGraph a = Gr a FEdge

-- | This is the \Sum_in in formula (2) of Louvain paper
type InWeightSum = Double
-- | This is the \Sum_tot in formula (2) of Louvain paper
type TotWeightSum = Double
newtype Community = Community { unCommunity ::  ([Node], InWeightSum, TotWeightSum) }

type CGrNode = Node
type CGrEdge = (InWeightSum, TotWeightSum)

type CGr = Gr Community CGrEdge


-- ALGORITHM

type Delta a b = Gr a b -> Node -> Community -> Double
-- | Delta Q function from Louvain paper (2).
delta :: Delta a b
delta gr n (Community (ns, inWeightSum, totWeightSum)) = 0.0

-- | One iteration step takes the graph and existing communities as a graph and
-- computes new community graph
iteration :: FGraph a -> CGr -> CGr
iteration gr cs = xdfsFoldWith suc' step cs (nodes gr) gr

-- TODO Remember to filter out empty Communities
-- | Step for one node. We try re-assign it to a neighbouring community, where
-- the increase of modularity for graph will be the largest
step :: CFunFold a FEdge CGr
step (p, v, l, s) cgr = cgr
  where
    mNc = nodeCommunity  v cgr
    ncs = nodeNeighbours v cgr

    -- We move node from community nc into ncs
    moves :: Maybe (LNode Community, [LNode Community])
    moves = case mNc of
      Nothing -> Nothing
      Just nc -> Just ( makeMove OutOf nc
                      , map (makeMove Into) ncs )

    makeMove :: Direction -> LNode Community -> LNode Community
    makeMove direction (cn, c) = (cn, moveNodeWithNeighbours (p <> s) v direction c)

    -- TODO Compute \Delta Q (gain of moving node v into Community C) which consists of:
    -- - Community WeightSum
    -- - sum of weights of links incident to nodes in C
    -- - sum of weights of links incident to node v
    -- - sum of weights of links from node v to nodes in C
    -- - sum of weights of all the links in the network
    -- So the Delta function takes as parameters:
    -- C, (edges from C in cgr), (edges from v in gr), (edges from v to C), (edges in gr)


-- COMMUNITY GRAPH FUNCTIONS

-- | 'Direction' when moving node 'Into'/'OutOf' community
data Direction = Into | OutOf

-- | Given 'Node' and 'Community' graph, find the 'LNode' of 'Community' which
-- contains the node
nodeCommunity :: Node -> CGr -> Maybe (LNode Community)
nodeCommunity n cgr = head (filter f $ labNodes cgr)
  where
    f :: (a, Community) -> Bool
    f (_, Community com) = n `elem` fst3 com

-- | Find 'LNode's of 'Community' graph neighbouring a given node
nodeNeighbours :: Node -> CGr -> [LNode Community]
nodeNeighbours n cgr =
  case nodeCommunity n cgr of
    Nothing      -> []
    Just (cn, _) -> mapMaybe (lnode cgr) (neighbors cgr cn)

-- | Find 'Ajd CGrEdge's of 'Community' graph neighbouring a given node
nodeLNeighbours :: Node -> CGr -> Adj CGrEdge
nodeLNeighbours n cgr =
  case nodeCommunity n cgr of
    Nothing      -> []
    Just (cn, _) -> lneighbors cgr cn

-- | Moves 'Node' in the 'Direction' of 'Community' and recomputes 'Community''s weights
moveNode :: FGraph a -> Node -> Direction -> Community -> Community
moveNode gr n direction c = moveNodeWithNeighbours lnNeighbors n direction c
  where
    lnNeighbors :: Adj FEdge
    lnNeighbors = lneighbors gr n

-- | Same as 'moveNode' above but with only node neighbours, not whole graph
moveNodeWithNeighbours :: Adj FEdge -> Node -> Direction -> Community -> Community
moveNodeWithNeighbours lnNeighbors n direction (Community (ns, inwsum, totwsum)) =
  Community (newNs, newInWsum, newTotWsum)
    where
      newNs :: [Node]
      newNs = case direction of
        Into  -> n:ns
        OutOf -> DL.delete n ns

      newInWsum, newTotWsum :: Double
      newInWsum  = inwsum  + directionN * sumN
      newTotWsum = totwsum + directionN * (sumN - sumNonCom)

      directionN :: Double
      directionN = case direction of
        Into  ->  1
        OutOf -> -1

      -- Update InWeightSum with connections between node and the community
      sumN :: InWeightSum
      sumN = sum $ map fst comNeighbors

      -- Update TotWeightSum, subtracting connections between node and community
      -- and adding connections of node to non-community
      sumNonCom :: TotWeightSum
      sumNonCom = sum $ map fst nonComNeighbors

      -- Node Adj Context
      comNeighbors    :: Adj FEdge
      comNeighbors    = filter (\ln -> snd ln `elem`    ns) lnNeighbors

      nonComNeighbors :: Adj FEdge
      nonComNeighbors = filter (\ln -> snd ln `notElem` ns) lnNeighbors


{-
-- | Moves 'Node' between two 'Community'ies, recomputing their weights
moveNode' :: FGraph a -> Node -> (Community, Community) -> (Community, Community)
moveNode' gr n cFromTo = moveNodeWithNeighbours' lnNeighbors n cFromTo
  where
    lnNeighbors :: Adj FEdge
    lnNeighbors = lneighbors gr n

-- | Same as 'moveNode' but with direct neighbours list
moveNodeWithNeighbours' :: Adj FEdge -> Node -> (Community, Community) -> (Community, Community)
moveNodeWithNeighbours' lnNeighbors n ((Community (fromNs, fromInWSum, fromTotWSum)), (Community (toNs, toInWSum, toTotWSum))) =
  (newFrom, newTo)
  where
    newFrom = Community (DL.delete n fromNs, newFromInWSum, newFromTotWSum)
    newTo = Community (n:toNs, newToInWSum, newToTotWSum)

    -- Node is removed, so we reduce the internal weight sum
    newFromInWSum = fromInWSum - sumFromN
    newFromTotWSum = fromTotWSum + sumFromN - sumFromNonN

    -- Node is added, so we increase the internal weight sum
    newToInWSum = toInWSum + sumToN
    newToTotWSum = toTotWSum + sumToN - sumToNonN

    sumFromN :: InWeightSum
    sumFromN = sum $ map fst fromComNeighbors
    sumFromNonN :: TotWeightSum
    sumFromNonN = sum $ map fst fromNonComNeighbors

    sumToN :: InWeightSum
    sumToN = sum $ map fst toComNeighbors
    sumToNonN :: TotWeightSum
    sumToNonN = sum $ map fst toNonComNeighbors

    fromComNeighbors :: Adj FEdge
    fromComNeighbors = filter (\ln -> snd ln `elem` fromNs) lnNeighbors
    fromNonComNeighbors :: Adj FEdge
    fromNonComNeighbors = filter (\ln -> snd ln `notElem` fromNs) lnNeighbors
    toComNeighbors :: Adj FEdge
    toComNeighbors = filter (\ln -> snd ln `elem` toNs) lnNeighbors
    toNonComNeighbors :: Adj FEdge
    toNonComNeighbors = filter (\ln -> snd ln `notElem` toNs) lnNeighbors
-}
