{-|
Module      : Data.Graph.Clustering.HLouvain
Description : Purely functional (Inductive) Louvain clustering
Copyright   : (c) Alexandre Delanoë, CNRS, 2020-Present
License     : AGPL + CECILL v3
Maintainer  : alexandre.delanoe+louvain@iscpif.fr
Stability   : experimental
Portability : POSIX

Clustering is not precise enough to work with (kmeans clustering has not
the same methodology than Louvain clustering fo instance, not speaking
about the use of graph or not that is not the point).

Here is a new semantic to change this situation.
glustering vs klustering

glustering: agglomerate according to glue rules (bottom up)
klustering: split       according to klue rules (top down)

-}

{-# LANGUAGE ConstrainedClassMethods #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE InstanceSigs      #-}
{-# LANGUAGE NoImplicitPrelude #-}

module Data.Graph.Clustering.HLouvain
  where

import Protolude
import Data.Set (Set)
import qualified Data.Set as Set
import Data.Ord (Down(..), compare)
import Data.List (foldl', maximumBy)
import Data.Graph.Inductive (DynGraph, Node, Graph)
import qualified Data.Graph.Inductive as G

------------------------------------------------------------------------
-- | Main Tools
------------------------------------------------------------------------
-- | Glustering (instead of clustering) since we are working on the glue
-- between nodes to create "glusters" of nodes: this API will enable the 
-- choice of "glue" (as a function) for the glustering.

glustering :: (DynGraph gr, Eq b)
           => gr a b
           -> [[Set Node]]
glustering g = map (map nodesD) $ glusteringD g

glusteringD :: (DynGraph gr, Eq b)
           => gr a b
           -> [[Dendogram]]
glusteringD g = map (glusteringC g) (comp g)

glusteringC :: DynGraph gr
           => gr a b
           -> [Node]
           -> [Dendogram]
glusteringC g ns = foldl' (\ds n -> insert g (toDendo g n) ds) [] ns

------------------------------------------------------------------------
type Component = [Node]

comp :: (DynGraph gr, Eq b) => gr a b -> [Component]
comp g' = map sortNodes cs
  where
    sortNodes ns = case head $ sortOn (Down . (G.deg g)) ns of
      Nothing -> []
      Just  n -> G.dfs [n] g  -- dfs for glustering, bfs for klustering
    cs = G.components g
    g  = G.undir g'

------------------------------------------------------------------------
-- | Definitions
------------------------------------------------------------------------

-- | A dendogram is a Tree whose leaf is one gram only and parents nodes
-- are glusters with score of glue to build it.

data Dendogram  = Gram Node
                | Dendogram { dends :: ![Dendogram] -- d-ends ends can be either glusters or klusters
                            , score :: !Double
                            }
  deriving (Show, Eq)

--------------------------------------------------------------------------
insert :: DynGraph gr
       => gr n e
       ->  Dendogram
       -> [Dendogram]
       -> [Dendogram]
insert g d ds = maximumBy (\a b -> compare (hasScore g a) (hasScore g b))
              $ insertD g d ds

insertD :: DynGraph gr
       => gr n e
       ->   Dendogram
       ->  [Dendogram]
       -> [[Dendogram]]
insertD _ d []     = [[d]]
insertD g d (x:xs) = [[inDendo g d x] ++ xs]
                  ++ (map (\y -> [x] ++ y) $ insertD g d xs)

--------------------------------------------------------------------------
toDendo :: DynGraph gr
        => gr n e
        -> Node
        -> Dendogram
toDendo g n = Dendogram [Gram n] s
  where
    s = modularity g (Set.singleton n)

toDendoD :: DynGraph gr
         => gr n e
         -> Dendogram
         -> Dendogram
toDendoD g d@(Gram _)        = Dendogram [d] (hasScore g d)
toDendoD _ d@(Dendogram _ _) = d

--------------------------------------------------------------------------
inDendo :: DynGraph gr
        => gr n e
        -> Dendogram
        -> Dendogram
        -> Dendogram

inDendo g d1@(Gram _) d2@(Gram _)   = Dendogram n' (hasScore g n')
  where
    ds = inDendo g d1 (toDendoD g d2)
    ss = hasScore g ds

    s1 = hasScore g d1
    s2 = hasScore g d2

    n' = [d1,d2]
{-
    n' | s1 + s2 > ss = [d1,d2]
       | otherwise    = [ds]
-}

inDendo g d@(Dendogram _ _) n@(Gram _) = inDendo g n d
inDendo g n@(Gram _) d@(Dendogram _ _) = inDendo g (toDendoD g n) d

inDendo g d1@(Dendogram _ s1) d2@(Dendogram _ s2) = Dendogram n' (hasScore g n')
  where
    ds = inDendo g d1 d2
    ss = hasScore g ds
    n' = [ds]
{-
    n' | s1 + s2 > ss = [d1,d2]
       | otherwise    = [ds]
-}
-------------------------------------------------------------------
nodesD :: Dendogram -> Set Node
nodesD (Gram n)         = Set.singleton n
nodesD (Dendogram [] _) = Set.empty
nodesD (Dendogram s  _) = Set.unions $ map nodesD s

deg :: Graph gr => gr a b -> Dendogram -> Int
deg g d = Set.size (neighbors g d)

neighbors :: Graph gr => gr a b -> Dendogram -> Set Node
neighbors g d = exclusion inNodes ouNodes
  where
    inNodes :: Set Node
    inNodes = nodesD d

    ouNodes :: Set Node
    ouNodes = Set.unions $ map (neighbors' g) (Set.toList inNodes)

    neighbors' :: Graph gr => gr a b -> Node -> Set Node
    neighbors' g' n' = Set.fromList $ G.neighbors g' n'

class HasScore a
  where
    hasScore :: DynGraph gr => gr n e -> a -> Double

instance HasScore Dendogram
  where
    hasScore :: DynGraph gr => gr a b -> Dendogram -> Double
    hasScore g m = modularity g (nodesD m)

instance HasScore [Dendogram]
  where
    hasScore g ds = modularity g (Set.unions $ map nodesD ds)

modulare :: DynGraph gr => gr a b -> Set Node -> Double
modulare = modularity

modularity :: DynGraph gr => gr a b -> Set Node -> Double
modularity gr ns = coverage - edgeDensity
  where
    coverage :: Double
    coverage  = sizeSubGraph / sizeAllGraph 
      where
          sizeSubGraph :: Double
          sizeSubGraph = fromIntegral ( G.size $ subgraph ns gr )

          sizeAllGraph :: Double
          sizeAllGraph = fromIntegral (G.size gr)

    edgeDensity :: Double
    edgeDensity = (sum (Set.map (\node -> (degree node) / links ) ns)) ** 2
      where
          degree :: Node -> Double
          degree node = fromIntegral (G.deg gr node)

          links :: Double
          links       = fromIntegral (2 * (G.size gr))

subgraph :: DynGraph gr => Set Node -> gr a b -> gr a b
subgraph ns = G.subgraph (Set.toList ns)

exclusion :: Ord a => Set a -> Set a -> Set a
exclusion a b = (Set.\\) b a


{-
Union Find algorithm could be used but Inductive Functional Graph
exactly have been created by Martin Erwig to avoid such style of
programming to enable clearer Graphs algorithms

import Control.Monad (mapM)
-- import Data.UnionFind.ST as ST
import Control.Monad.Trans.UnionFind as ST
-- import Control.Monad.ST  as ST
import Control.Monad.State as ST

    data UnionFindT p m a
    runUnionFind :: Monad m => UnionFindT p m a -> m a
    data Point a
    fresh :: Monad m => p -> UnionFindT p m (Point p)
    repr :: Monad m => Point p -> UnionFindT p m (Point p)
    descriptor :: Monad m => Point p -> UnionFindT p m p
    union :: Monad m => Point p -> Point p -> UnionFindT p m ()
    equivalent :: Monad m => Point p -> Point p -> UnionFindT p m Bool

unionFind :: IO [Int]
unionFind = ST.runUnionFind $ do
  ns <- mapM ST.fresh ([1..10] :: [Int])
  ST.union (ns !! 1) (ns !! 2)
  ST.union (ns !! 3) (ns !! 4)
  ST.union (ns !! 5) (ns !! 6)
  t1 <- ST.equivalent (ns !! 1) (ns !! 2)
  ST.lift $ putStrLn $ show t1
  t2 <- ST.equivalent (ns !! 4) (ns !! 5)
  n  <- mapM ST.repr ns
  n' <- mapM ST.descriptor n
  pure n'
-}

